/**
 * Pikku esimerkki POSIX-säikeistä.
 *
 * Mihin saatettaisiin tarvita säikeitä? Tässä on esimerkki
 * rinnakkaistuvasta tehtävästä, jonka seinäkelloaika saadaan
 * lyhyemmäksi rinnakkaislaskennan avulla.
 *
 * Muita tarpeita? Tausta-ajot, esim. grafiikkaohjelman
 * efektialgoritmit tai nettiselaimen välilehdet tai
 * Download-ominaisuus. Käyttöliittymäsäiettä voisi klikkailla, vaikka
 * ohjelma tekisi samaan aikaan taustatöitä eri paikoissa koodia ja
 * dataa.
 *
 *  HUOM: Esimerkkien lyhentämiseksi näissä ei tarkisteta,
 *  onnistuivatko operaatiot. Oikeassa ohjelmassa AINA tarkistettava
 *  esim. onnistuiko säikeen luonti vai ei, ja käsiteltävä
 *  epäonnistuminen!!
 */

#include<stdint.h>
#include<stdio.h>
#include<inttypes.h>
#include<stdlib.h>
#include<pthread.h>

#define TAULUN_KOKO 60000000
#define MAX_SAIKEITA 16

typedef struct {
    int64_t *taulukko;
    int64_t alkuindeksi;
    int64_t loppuind_plusyks;
} saikeen_tiedot;

void tayta_perakkaisilla(int64_t *t, int64_t a, int64_t b){
    for(int64_t i = a; i<b; i++){
        t[i] = i;
    }
}

void * saikeen_koodi(void *tied) {
    int i;
    int64_t *t = ((saikeen_tiedot*)tied)->taulukko;
    int64_t a = ((saikeen_tiedot*)tied)->alkuindeksi;
    int64_t b = ((saikeen_tiedot*)tied)->loppuind_plusyks;
    tayta_perakkaisilla(t,a,b);
    /*
    for(int64_t i = a; i<b; i++){
        t[i] = i;
        }*/
    return NULL;
}

void tayta_perakkaisilla_saikeissa(int64_t *t, int64_t a, int64_t b, size_t nthr){
    if ((b-a) < nthr) nthr = 1;
    if (nthr >= MAX_SAIKEITA) nthr = MAX_SAIKEITA;
    pthread_t th[MAX_SAIKEITA];
    saikeen_tiedot tied[MAX_SAIKEITA];

    int64_t vali = (b-a)/nthr;
    int64_t jakopiste = a;
    for(int i=0;i<nthr;i++){
        tied[i].taulukko=t;
        tied[i].alkuindeksi=jakopiste;
        jakopiste += vali;
        tied[i].loppuind_plusyks=((jakopiste<=b)?jakopiste:b);
        printf("a=%lld b=%lld\n",tied[i].alkuindeksi,tied[i].loppuind_plusyks);
    }
    for(size_t i=0;i<nthr;i++){
        pthread_create(&th[i], NULL, saikeen_koodi, &tied[i]);
    }
    for(size_t i=0;i<nthr;i++){
        pthread_join(th[i], NULL);
    }
}

int main(int argc, char **argv){
    int64_t *taulukko = malloc(TAULUN_KOKO*sizeof(int64_t));
    //tayta_perakkaisilla(taulukko, 0, TAULUN_KOKO);
    size_t nsaikeita;
    if (argc < 2){
        nsaikeita = 1;
    } else {
        nsaikeita = atoi(argv[1]);
    }
    
    tayta_perakkaisilla_saikeissa(taulukko, 0, TAULUN_KOKO,nsaikeita);

    // Köyhän miehen testi.. tokko toimi sinne päinkään (pistokoe):
    for(size_t ind=0;ind<TAULUN_KOKO;ind += (TAULUN_KOKO/7+1))
        printf("taulukko[%d]==%" PRId64 "\n", ind, taulukko[ind]);
    printf("taulukko[%d]==%" PRId64 "\n", TAULUN_KOKO-1, taulukko[TAULUN_KOKO-1]);
    return 0;
}
